@@ -10,14 +10,13 @@ import pathlib
 import re
 from collections import defaultdict
 from copy import copy
-from typing import List
+from typing import TYPE_CHECKING, List
 
 from fhirpath.enums import FHIR_VERSION
 from fhirpath.interfaces import IStorage
 from fhirpath.storage import MemoryStorage
 from fhirpath.utils import reraise
 
-
 logger = logging.getLogger("fhirpath.fhrspec")
 
 # allow to skip some profiles by matching against their url (used while WiP)
@@ -26,8 +25,7 @@ HTTP_URL = re.compile(r"^https?://", re.IGNORECASE)
 
 
 class FHIRSearchSpec(object):
-    """https://www.hl7.org/fhir/searchparameter-registry.html
-    """
+    """https://www.hl7.org/fhir/searchparameter-registry.html"""
 
     def __init__(
         self, source: pathlib.Path, fhir_release: FHIR_VERSION, storage: MemoryStorage
@@ -76,11 +74,13 @@ class FHIRSearchSpec(object):
         """ """
         storage = self.storage.get(self.fhir_release.name)
         base_resource_params = storage.get("Resource")
+        base_domain_resource_params = storage.get("DomainResource")
 
         for resource_type in storage:
             if resource_type in ("Resource", "DomainResource"):
                 continue
             storage.get(resource_type) + base_resource_params
+            storage.get(resource_type) + base_domain_resource_params
 
     @property
     def jsonfilename(self):
@@ -91,6 +91,20 @@ class FHIRSearchSpec(object):
 class SearchParameterDefinition(object):
     """ """
 
+    if TYPE_CHECKING:
+        spec = None
+        name: None
+        code: None
+        expression_map: None
+        type: None
+        modifier: None
+        comparator: None
+        target: None
+        xpath: None
+        multiple_or: None
+        multiple_and: None
+        component: None
+
     __slots__ = (
         "spec",
         "name",
@@ -152,6 +166,19 @@ class SearchParameterDefinition(object):
 class SearchParameter(object):
     """ """
 
+    if TYPE_CHECKING:
+        name: None
+        code: None
+        expression: None
+        type: None
+        modifier: None
+        comparator: None
+        target: None
+        xpath: None
+        multiple_or: None
+        multiple_and: None
+        component: None
+
     __slots__ = (
         "name",
         "code",
