@@ -8,16 +8,24 @@ import re
 import sys
 import uuid
 from importlib import import_module
+from inspect import signature
 from types import ModuleType
-from typing import Any
-from typing import List
-from typing import Match
-from typing import Optional
-from typing import Pattern
-from typing import Text
-from typing import cast
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    List,
+    Match,
+    Optional,
+    Pattern,
+    Text,
+    Type,
+    Union,
+    cast,
+)
 
 import pkg_resources
+from pydantic.validators import bool_validator
 from yarl import URL
 from zope.interface import implementer
 
@@ -25,10 +33,19 @@ from fhirpath.thirdparty import Proxy
 
 from .enums import FHIR_VERSION
 from .interfaces import IPathInfoContext
-from .storage import FHIR_RESOURCE_CLASS_STORAGE
-from .storage import PATH_INFO_STORAGE
+from .storage import FHIR_RESOURCE_CLASS_STORAGE, PATH_INFO_STORAGE
 from .types import PrimitiveDataTypes
 
+if TYPE_CHECKING:
+    from fhir.resources.fhirabstractmodel import FHIRAbstractModel  # noqa: F401
+    from fhir.resources.fhirtypes import (  # noqa: F401
+        AbstractBaseType,
+        AbstractType,
+        Primitive,
+    )
+    from pydantic.fields import ModelField  # noqa: F401
+    from pydantic.main import BaseConfig  # noqa: F401
+
 
 __author__ = "Md Nazrul Islam <email2nazrul@gmail.com>"
 
@@ -78,7 +95,7 @@ def force_bytes(
             return string.decode("utf8", errors).encode(encoding, errors)
 
     if not isinstance(string, str):
-        return string
+        string = str(string)
 
     return string.encode(encoding, errors)
 
@@ -91,7 +108,7 @@ def import_string(dotted_path: Text) -> type:
         module_path, class_name = dotted_path.rsplit(".", 1)
     except (ValueError, AttributeError):
         msg = f"{dotted_path} doesnt look like a module path"
-        reraise(ImportError, msg)
+        return reraise(ImportError, msg)
 
     module: ModuleType = import_module(module_path)
     cls: type
@@ -99,7 +116,7 @@ def import_string(dotted_path: Text) -> type:
         cls = getattr(module, class_name)
     except AttributeError:
         msg = f'Module "{module_path}" does not define a "{class_name}" attribute/class'
-        reraise(ImportError, msg)
+        return reraise(ImportError, msg)
     return cls
 
 
@@ -175,7 +192,7 @@ def lookup_fhir_class_path(
 
     prime_module_type: ModuleType = import_module(prime_module_path)
 
-    for importer, module_name, ispkg in pkgutil.walk_packages(
+    for _importer, module_name, ispkg in pkgutil.walk_packages(
         prime_module_type.__path__,  # type: ignore
         prime_module_type.__name__ + ".",
         onerror=lambda x: None,
@@ -185,7 +202,7 @@ def lookup_fhir_class_path(
 
         module_type: ModuleType = import_module(module_name)
 
-        for klass_name, klass in inspect.getmembers(module_type, inspect.isclass):
+        for klass_name, _klass in inspect.getmembers(module_type, inspect.isclass):
 
             if klass_name == resource_type:
                 storage.insert(resource_type, f"{module_name}.{resource_type}")
@@ -195,16 +212,25 @@ def lookup_fhir_class_path(
 
 def lookup_fhir_class(
     resource_type: Text, fhir_release: FHIR_VERSION = FHIR_VERSION.DEFAULT
-):  # noqa: E999
-    klass_path: Optional[Text] = lookup_fhir_class_path(
-        resource_type, True, fhir_release
-    )
-    klass: type = import_string(cast(Text, klass_path))
+) -> Type["FHIRAbstractModel"]:  # noqa: E999
+    factory_paths: List[str] = ["fhir", "resources"]
+    if (
+        FHIR_VERSION["DEFAULT"].value != fhir_release.name
+        and fhir_release != FHIR_VERSION.DEFAULT
+    ):
+        factory_paths.append(fhir_release.name)
+    factory_paths.append("get_fhir_model_class")
+
+    factory: type = import_string(".".join(factory_paths))
+    try:
+        klass = factory(resource_type)
+    except KeyError:
+        raise LookupError(f"{resource_type} is not a valid FHIR class")
     return klass
 
 
 CONTAINS_PY_PACKAGE: Pattern = re.compile(
-    r"^\$\{(?P<package_name>[0-9a-z._]+)\}", re.IGNORECASE
+    r"^\${(?P<package_name>[0-9a-z._]+)\}", re.IGNORECASE
 )
 
 
@@ -230,7 +256,7 @@ def expand_path(path_: Text) -> Text:
             msg = "Invalid package `{0}`! as provided in {1}".format(
                 package_name, path_
             )
-            reraise(LookupError, msg)
+            return reraise(LookupError, msg)
 
     else:
         real_path = path_
@@ -244,7 +270,7 @@ def expand_path(path_: Text) -> Text:
 def proxy(obj):
     """Making proxy of any object"""
     try:
-        return getattr(obj, "__proxy__")()
+        return obj.__proxy__()
     except AttributeError:
         # trying to make ourself
         p_obj = Proxy()
@@ -274,6 +300,7 @@ class EmptyPathInfoContext:
         self.type_class = None
         self.optional = None
         self.multiple = None
+        self.type_is_primitive = None
 
 
 EMPTY_PATH_INFO_CONTEXT = EmptyPathInfoContext()
@@ -285,30 +312,42 @@ class PathInfoContext:
 
     def __init__(
         self,
-        path,
-        fhir_release,
-        prop_name,
-        prop_original,
-        type_name,
-        type_class,
-        optional,
-        multiple,
+        path: str,
+        fhir_release: FHIR_VERSION,
+        prop_name: str,
+        prop_original: str,
+        type_name: str,
+        type_class: Union[bool, "AbstractBaseType", "AbstractType", "Primitive"],
+        type_field: "ModelField",
+        type_model_config: Type["BaseConfig"],
+        optional: bool,
+        multiple: bool,
+        type_is_primitive: bool,
+        resource_type: str,
     ):
         """ """
-        self._parent = None
-        self._children = list()
-        self._path = path
+        self._parent: Optional[str] = None
+        self._children: List[str] = list()
+        self._path: str = path
 
-        self.fhir_release = fhir_release
-        self.prop_name = prop_name
-        self.prop_original = prop_original
-        self.type_name = type_name
-        self.type_class = type_class
-        self.optional = optional
-        self.multiple = multiple
+        self.fhir_release: FHIR_VERSION = fhir_release
+        self.prop_name: str = prop_name
+        self.prop_original: str = prop_original
+        self.type_name: str = type_name
+        self.type_class: Union[
+            bool, "AbstractBaseType", "AbstractType", "Primitive"
+        ] = type_class
+        self.type_field: "ModelField" = type_field
+        self.type_model_config: Type["BaseConfig"] = type_model_config
+        self.optional: bool = optional
+        self.multiple: bool = multiple
+        self.type_is_primitive: bool = type_is_primitive
+        self.resource_type: str = resource_type
 
     @classmethod
-    def context_from_path(cls, pathname: Text, fhir_release: FHIR_VERSION):
+    def context_from_path(
+        cls, pathname: Text, fhir_release: FHIR_VERSION
+    ) -> Union["PathInfoContext", "EmptyPathInfoContext"]:
         """ """
         if pathname == "*":
             return EMPTY_PATH_INFO_CONTEXT
@@ -322,52 +361,69 @@ class PathInfoContext:
             return storage.get(pathname)
 
         parts = pathname.split(".")
-        model_path = lookup_fhir_class_path(parts[0], fhir_release=fhir_release)
-        model_type: type = import_string(cast(Text, model_path))
+        resource_type = parts[0]
+        model_path = lookup_fhir_class_path(resource_type, fhir_release=fhir_release)
+        model_class: Type["FHIRAbstractModel"] = cast(
+            Type["FHIRAbstractModel"], import_string(cast(Text, model_path))
+        )
         new_path: Text = parts[0]
-        context = None
+        context: Optional["PathInfoContext"] = None
 
         for index, part in enumerate(parts[1:], 1):
 
             new_path = "{0}.{1}".format(new_path, part)
             if storage.exists(new_path):
                 context = storage.get(new_path)
+                if TYPE_CHECKING:
+                    assert context
                 if context.type_name in PrimitiveDataTypes:
                     if (index + 1) < len(parts):
                         raise ValueError("Invalid path {0}".format(pathname))
                     break
                 else:
-                    model_type = context.type_class
+                    klass = context.type_class
+                    model_class = lookup_fhir_class(
+                        klass.__resource_type__,  # type: ignore
+                        FHIR_VERSION[klass.__fhir_release__],  # type: ignore
+                    )
                     continue
 
-            for (
-                name,
-                jsname,
-                typ,
-                typ_name,
-                is_list,
-                of_many,
-                not_optional,
-            ) in model_type().elementProperties():
+            for field in model_class.element_properties():
 
-                if part != jsname:
+                if part != field.alias:
                     continue
-                if typ_name in PrimitiveDataTypes:
-                    type_class = PrimitiveDataTypes.get(typ_name)
-                    is_primitive = True
-                else:
-                    type_class = typ
+                type_model_config = model_class.__config__
+                multiple = str(field.outer_type_)[:12] == "typing.List["
+                if getattr(field.type_, "__resource_type__", None):
+                    # AbstractModelType
+                    model_class = lookup_fhir_class(
+                        field.type_.__resource_type__,
+                        FHIR_VERSION[field.type_.__fhir_release__],
+                    )
+                    type_name = field.type_.__resource_type__
                     is_primitive = False
+                else:
+                    is_primitive = True
+                    # Primitive
+                    type_name = getattr(field.type_, "__visit_name__", None)
+                    if type_name is None and field.type_ == bool:
+                        type_name = "boolean"
+                    if type_name is None:
+                        raise NotImplementedError
 
                 context = cls(
                     new_path,
                     fhir_release=fhir_release,
-                    prop_name=name,
-                    prop_original=jsname,
-                    type_name=typ_name,
-                    type_class=type_class,
-                    optional=(not not_optional),
-                    multiple=is_list,
+                    prop_name=field.name,
+                    prop_original=field.alias,
+                    type_name=type_name,
+                    type_class=field.type_,
+                    type_field=field,
+                    type_model_config=type_model_config,
+                    optional=(not field.required),
+                    multiple=multiple,
+                    type_is_primitive=is_primitive,
+                    resource_type=resource_type,
                 )
                 if index > 1:
                     context.parent = ".".join(new_path.split(".")[:-1])
@@ -377,15 +433,16 @@ class PathInfoContext:
                     parent_context.add_child(new_path)  # type: ignore
 
                 storage.insert(new_path, context)
-                if not is_primitive:
-                    model_type = cast(type, type_class)
-                else:
+
+                if is_primitive:
                     if (index + 1) < len(parts):
                         raise ValueError("Invalid path {0}".format(pathname))
                     break
             # important! even context is None, that means not valid path (part)
             if context is None:
                 break
+        if TYPE_CHECKING:
+            assert context
         return context
 
     def __proxy__(self):
@@ -396,9 +453,13 @@ class PathInfoContext:
         """ """
         self._parent = dotted_path
 
-    def _get_parent(self):
+    def _get_parent(self) -> "PathInfoContext":
         """ """
-        return PathInfoContext.context_from_path(self._parent, self.fhir_release)
+        assert self._parent
+        parent = PathInfoContext.context_from_path(self._parent, self.fhir_release)
+        if TYPE_CHECKING:
+            assert isinstance(parent, PathInfoContext)
+        return parent
 
     parent = property(_get_parent, _set_parent)
 
@@ -436,6 +497,23 @@ class PathInfoContext:
         """ """
         return str(self._path)
 
+    def validate_value(self, value):
+        """``pydantic`` way to validate value"""
+        if self.type_class == bool:
+            return bool_validator(value)
+        for validator in self.type_class.__get_validators__():
+            sig = signature(validator)
+            args = list(sig.parameters.keys())
+            if len(args) == 1:
+                value = validator(value)
+            elif len(args) == 2:
+                value = validator(value, self.type_field)
+            elif len(args) == 3:
+                value = validator(value, self.type_field, self.type_model_config)
+            else:
+                raise NotImplementedError
+        return value
+
 
 class PathInfoContextProxy(Proxy):
     """ """
@@ -449,63 +527,70 @@ class PathInfoContextProxy(Proxy):
 class BundleWrapper:
     """ """
 
-    def __init__(self, engine, result, url: URL, bundle_type="searchset"):
+    def __init__(
+        self, engine, result, includes: List, url: URL, bundle_type="searchset"
+    ):
         """ """
         self.fhir_version = engine.fhir_release
-        self.bundle = lookup_fhir_class("Bundle", fhir_release=self.fhir_version)()
-        self.bundle.id = str(uuid.uuid4())
+        self.bundle_model = lookup_fhir_class("Bundle", fhir_release=self.fhir_version)
+        self.data: Dict[str, Any] = dict()
+        self.data["id"] = str(uuid.uuid4())
 
-        self.bundle.meta = lookup_fhir_class("Meta", fhir_release=self.fhir_version)()
+        self.data["meta"] = {"lastUpdated": datetime.datetime.utcnow()}
 
-        fhir_dt = lookup_fhir_class("FHIRDate", fhir_release=self.fhir_version)()
-        fhir_dt.date = datetime.datetime.now()
-        self.bundle.meta.lastModified = fhir_dt
-
-        self.bundle.type = bundle_type
-        self.bundle.total = result.header.total
+        self.data["type"] = bundle_type
+        # our pagination is based main query result.
+        # fixme: still issue for _has chaining
+        self.data["total"] = result.header.total
 
+        # attach main results
         self.attach_entry(result, "match")
 
+        # attach included results
+        for _include in includes:
+            self.attach_entry(_include, "include")
+
         self.attach_links(url, len(result.body))
 
     def attach_entry(self, result, mode="match"):
         """ """
-        if not self.bundle.entry:
-            self.bundle.entry = list()
+        if "entry" not in self.data:
+            self.data["entry"] = list()
 
         for row in result.body:
-            entry = row[0]
-            resource = lookup_fhir_class(
-                entry["resourceType"], fhir_release=self.fhir_version
-            )(entry)
+            resource = row[0]
+            if isinstance(resource, dict):
+                resource_id = resource["id"]
+                resource_type = resource["resourceType"]
+            else:
+                resource_id = resource.id
+                resource_type = resource.resource_type
+            # entry = BundleEntry
+            entry = dict()
+            entry["fullUrl"] = "{0}/{1}".format(resource_type, resource_id)
+            entry["resource"] = resource
+            # search = BundleEntrySearch
+            search = {"mode": mode}
+            entry["search"] = search
 
-            item = lookup_fhir_class("BundleEntry", fhir_release=self.fhir_version)()
-            item.fullUrl = "{0}/{1}".format(resource.resource_type, resource.id)
-            item.resource = resource
-
-            item.search = lookup_fhir_class(
-                "BundleEntrySearch", fhir_release=self.fhir_version
-            )()
-            item.search.mode = mode
-
-            self.bundle.entry.append(item)
+            self.data["entry"].append(entry)
 
     def attach_links(self, url, entries_count):
         """ """
         container = list()
 
         _max_count = int(url.query.get("_count", 100))
-        _total_results = self.bundle.total
+        _total_results = self.data["total"]
         _current_offset = int(url.query.get("search-offset", 0))
         url_params = {}
 
         container.append(self.make_link("self", url))
 
-        # let's pagitionation here
+        # let's pagination here
         if _total_results > _max_count:
             # Yes pagination is required!
             url_params["_count"] = _max_count
-            url_params["search-id"] = self.bundle.id
+            url_params["search-id"] = self.data["id"]
 
             # first page
             if _current_offset != 0:
@@ -533,28 +618,27 @@ class BundleWrapper:
                 url_params["search-offset"] = last_offset
                 container.append(self.make_link("last", url, url_params))
 
-        self.bundle.link = container
+        self.data["link"] = container
 
-    def make_link(self, relation, url, params={}):
+    def make_link(self, relation, url, params=None):
         """ """
-        item = lookup_fhir_class("BundleLink", fhir_release=self.fhir_version)()
-        item.relation = relation
+        params = params or {}
+        # link = BundleLink
+        link = {"relation": relation}
         # fix: params
         existing_params = url.query.copy()
         for key in params:
             existing_params[key] = params[key]
 
         new_url = url.with_query(existing_params)
-        item.url = str(new_url)
+        link["url"] = str(new_url)
 
-        return item
+        return link
 
     def __call__(self):
         """ """
-        # Validation purpose
-        self.as_json()
-        return self.bundle
+        return self.bundle_model.parse_obj(self.data)
 
-    def as_json(self):
+    def json(self):
         """ """
-        return self.bundle.as_json()
+        return self.__call__().json()
